/* $Id: $ */

/* Copyright (C) 1999 Petr Novak, Charles University, Prague.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
   libterm.sim                            V0.1    (Cim 3.29 / Win32)

   library functions implementation for the class terminal

   by Petr Novak 1999
*/



#include <windows.h>  
#include <stdio.h>
#include <string.h>

/*********************************  data ************************************/

HANDLE hStdout = INVALID_HANDLE_VALUE; 
HANDLE hStdin; 

POINT cursorPosition;
CONSOLE_SCREEN_BUFFER_INFO screenInfo; // for getting: size, atribute
CONSOLE_CURSOR_INFO curinfo;           // this is usefull for visibility & size
WORD currAttr;
BOOL printAll;         // if TRUE all char are printed visible
DWORD modeProcessed;   // mode processing controll characters (default)
DWORD modeAll;         // mode displaying all characters
DWORD oldInputMode;    // mode for restoring     
BOOL listenCtrls = FALSE, listenRelease = FALSE;
COORD lastPos;         // for position of the cursor
BOOL flushed;          // true, if input was flushed
BOOL mouseEnabled;     // true, if mouse events are enabled
BOOL insertMode;       // actual mode
DWORD curSize;         // original cursor size before opening terminal

/********************************* opened control ***************************/

BOOL isOpened = FALSE;

void termError (char * msg);
void sysError (char * msg);

inline void testOpened (char * fnName) 
{
   char errmes [35];
   strcat (strcpy (errmes, fnName), ": used without \"open\"");
   if (!isOpened) __rerror (errmes);
}

/********************************* functions ********************************/

void setModeAll (char enable)
{
  if (enable) {
    SetConsoleMode (hStdout, modeAll);
  }
  else {
    SetConsoleMode (hStdout, modeProcessed);
  }
}

void EprintAll (char enable)
{
  printAll = enable;
  setModeAll (enable);
}

void setMouse (char on)
{
  DWORD mode;
  testOpened ("setMouse");
  mouseEnabled = on;
  GetConsoleMode (hStdin, &mode);
  /* Note:
     next condition is here because of undocumented behaviour of setmode:
     as many times you call setmode (mouse disable) so many times you have
     to call setmode (mouse enable) to visualise caret. With this cond you
     can call setMouse (false) as many times you want.
   */
  if (!on && !(mode & ENABLE_MOUSE_INPUT)) return; 

  if (on) {
    mode |= ENABLE_MOUSE_INPUT;
  }
  else {
    mode &= ~ENABLE_MOUSE_INPUT;
  }
  if (! SetConsoleMode(hStdin, mode) )
    sysError("SetConsoleMode");
}

void autoWrap(char on)
{
  DWORD mode;
  testOpened ("autoWrap");

  GetConsoleMode (hStdout, &mode);
  if (on) {
    mode          |= ENABLE_WRAP_AT_EOL_OUTPUT;
    modeProcessed |= ENABLE_WRAP_AT_EOL_OUTPUT;
    modeAll       |= ENABLE_WRAP_AT_EOL_OUTPUT;
  }
  else {
    mode          &= ~ENABLE_WRAP_AT_EOL_OUTPUT;
    modeProcessed &= ~ENABLE_WRAP_AT_EOL_OUTPUT;
    modeAll       &= ~ENABLE_WRAP_AT_EOL_OUTPUT;
  }
  SetConsoleMode (hStdout, mode);
}

void termError (char * message)
{
  char errmes [75];
  strcat (strcpy (errmes, "Terminal: "), message);
  if (isOpened) SetConsoleMode (hStdout, modeProcessed);
  __rerror (errmes);
}

void sysError (char * message)
{
  char errmes [75];
  strcat (strcpy (errmes, "Terminal: "), message);
  if (isOpened) SetConsoleMode (hStdout, modeProcessed);
  __rsystemerror (errmes);
}

long wAND(long i, long j)
{
  return (i & j);
}

long wOR(long i, long j)
{
  return (i | j);
}

long wSHL(long i, long j)
{
  return (i << j);
}

long wSHR(long i, long j)
{
  return (i >> j);
}

char isSet (unsigned char atr, long bit)
{
  return (atr & bit) != 0;
}

long EHEIGHT()
{
  return screenInfo.dwSize.Y;
}

long EWIDTH()
{
  return screenInfo.dwSize.X;
}

void EshowCursor ()
{
  testOpened ("showCursor");
//  ShowCursor (TRUE);
  curinfo.bVisible = TRUE;
  SetConsoleCursorInfo (hStdout, &curinfo);
}
void EhideCursor ()
{
  testOpened ("hideCursor");
//  ShowCursor (FALSE);
  curinfo.bVisible = FALSE;
  SetConsoleCursorInfo (hStdout, &curinfo);
}

void setInsert (char on)
{
  testOpened ("hideCursor");
  insertMode = on;
  curinfo.dwSize = (on ? 15 : 99);
  SetConsoleCursorInfo (hStdout, &curinfo);
}

long ECURLIN()
{
  testOpened ("curLin");
//  GetCursorPos (&cursorPosition);
//  return cursorPosition.y;
  GetConsoleScreenBufferInfo (hStdout, &screenInfo);
  return screenInfo.dwCursorPosition.Y;
}

long ECURPOS()
{
  testOpened ("curPos");
//  GetCursorPos (&cursorPosition);
//  return cursorPosition.x;
  GetConsoleScreenBufferInfo (hStdout, &screenInfo);
  return screenInfo.dwCursorPosition.X;
}
void EUPDPOS(long lin, long pos)
{
  testOpened ("setPos");
//  SetCursorPos (pos, lin);
  lastPos.X = pos >= EWIDTH  () ? EWIDTH  () - 1 : pos;
  lastPos.Y = lin >= EHEIGHT () ? EHEIGHT () - 1 : lin;
  SetConsoleCursorPosition (hStdout, lastPos);
}

#define SetCursorPos(lin, pos) EUPDPOS ((pos), (lin))

inline void setAttr (long a)
{
  if (a != currAttr) {
    SetConsoleTextAttribute (hStdout, a);
    currAttr = a;
  }
}

void eraseLines(int from, int to, long attr)
{
    DWORD cWritten; 
    BOOL fSuccess; 
    COORD coord = {0, from}; 
    DWORD nLength = (to - from + 1) * EWIDTH ();

    static CHAR const chFillChar = ' '; 

    fSuccess = FillConsoleOutputCharacter(
        hStdout,          /* screen buffer handle     */
        chFillChar,       /* fill with spaces         */
        nLength,          /* number of cells to fill  */
        coord,            /* first cell to write to   */

        &cWritten);       /* actual number written to */
    if (! fSuccess)
        sysError("FillConsoleOutputCharacter");

    fSuccess = FillConsoleOutputAttribute(
        hStdout,          /* screen buffer handle     */
        attr,             /* color to fill with       */

        nLength,           /* number of cells to fill  */
        coord,            /* first cell to write to   */
        &cWritten);       /* actual number written to */

    if (! fSuccess)
        sysError("FillConsoleOutputAttribute");
}


void EerasePage(long c)
{
  testOpened ("erasePage");
  setAttr (c);
  eraseLines (0, EHEIGHT () - 1, c);
  SetCursorPos (0, 0);
}

void EeraseLine(long c)
{
  testOpened ("eraseLine");
  setAttr (c);
  eraseLines (ECURLIN (), ECURLIN (), c);
}

void EOUTCHR(char c, long a)
{
  DWORD cWritten;
  static char charBuff [] = " ";

  testOpened ("outChar");
  setAttr (a);
  charBuff [0] = c;
  WriteFile( 
            hStdout,              /* output handle  */ 
            charBuff,             /* prompt string  */ 
            1,                    /* string length  */ 
            &cWritten,            /* bytes written  */ 
            NULL);                /* not overlapped */ 
}

void EOUTTXT(char *t, long w, long a) 
{
  static DWORD cWritten;

  testOpened ("outText");
  setAttr (a);
  WriteFile(
            hStdout,              /* output handle  */
            t,                    /* prompt string  */
            w,                    /* string length  */
            &cWritten,            /* bytes written  */
            NULL);                /* not overlapped */
}

void ENEWLIN(long a)
{
  testOpened ("newLine");
  SetConsoleMode (hStdout, modeProcessed);
  EOUTCHR('\n', a);
  EprintAll (printAll);
}

/* unfortunately, this is working only under NT */
void EBEEP ()
{
  testOpened ("beep");
  SetConsoleMode (hStdout, modeProcessed);
  EOUTCHR('\007', currAttr);
  EprintAll (printAll);
}

void Esleep (long ms)
{
  Sleep (ms);
}

void EGETTXT(long lin, long pos, long width, char *t)
{
    static SMALL_RECT srctReadRect; 
    static COORD coordBufSize; 
    static COORD coordBufCoord; 
    static BOOL fSuccess; 
    int i, length;

    testOpened ("getText");

    length = EWIDTH () - pos;
    length = (length < width ? length : width);
    if (length > strlen (t)) termError ("small buffer for getText");

    {
    CHAR_INFO chiBuffer[length];

    srctReadRect.Top = lin;     
    srctReadRect.Left = pos; 
    srctReadRect.Bottom = lin;  
    srctReadRect.Right = pos + length - 1; 
 
    coordBufSize.Y = 1; 
    coordBufSize.X = length; 
 
    /* The top left destination cell of the temporary buffer 
     * row 0, col 0. 
     */ 
    coordBufCoord.X = 0; 
    coordBufCoord.Y = 0; 

    fSuccess = ReadConsoleOutput( 
       hStdout,        /* screen buffer to read from       */ 
       chiBuffer,      /* buffer to copy into              */ 
       coordBufSize,   /* col-row size of chiBuffer        */ 
       coordBufCoord,  /* top left dest. cell in chiBuffer */ 
       &srctReadRect); /* screen buffer source rectangle   */ 
    for (i = 0; i < length; i++) t[i] = chiBuffer[i].Char.AsciiChar;
    }
}

typedef struct /* this structure represents simula object keyPkt */
    {
        unsigned char  map;
        unsigned char  chr;
        long           fkey;
        unsigned char  c0;
        unsigned char  c1;
        unsigned char  c2;
        long           x, y;
    } keypkt;

char const moveKey[] =
    { 45, 46, 36, 35, 33, 34, 37, 38, 40, 39, 0 }; // virtualKeyCode
char const mapMoveKey [] =
    {  3,127, 15, 26, 30, 31,  5, 28, 14,  4, 0 }; // simula map

long const invisibleKey [] =
    { 0, 13, 16, 17, 18, 20, 144, 145, -1 };

/*
  documentation for keyin:

  map - mapped key, like in keyboard.ins
  chr - scan code (exact key on the keyboard)
  fkey - function key according to keyboard.ins and many others
  c0, c1 - control keys (each bit signs some control key)
  c2 - key release (if enabled by listenReleaseKeys)
  
  - you can test the behaviour of control keys using listenControllKeys
  - I suggest to do some tests before programming nontrivial key-combination
*/

void KEYIN(char wait, void * pkt)
{
  /* this routin uses lo-level input winapi */
  /* to recognize control keys and */
  /* to be able to handle mouse events in the future */

  static INPUT_RECORD irInBuf[128];
  static DWORD nextEvent = 0;
  static DWORD cNumRead  = 0;

  KEY_EVENT_RECORD * keyEvent;
  MOUSE_EVENT_RECORD * mouseEvent;
  DWORD numEvents, i;
  keypkt * rk = pkt;

  testOpened ("inkey, ifkey");

  if (flushed) {                        // flushedInput was called
    flushed = FALSE;
    nextEvent = cNumRead  = 0;
  }

  // EprintAll (FALSE); printf ("\nkeyin\n");
  while (1) { 
 
    // if user called 'ifchar' and no keys are in the input:
    GetNumberOfConsoleInputEvents (hStdin, &numEvents);
    if (cNumRead == nextEvent && numEvents == 0 && wait == FALSE) {
      rk -> map = rk -> chr = rk -> fkey = rk -> c0 = rk -> c1 = 0;
      return;
    }

    /* Wait for the events. */
      
    if (cNumRead == nextEvent) {
      nextEvent = 0;
      // printf ("readinginput\n");
      if (! ReadConsoleInput( 
                hStdin,      /* input buffer handle    */ 

                irInBuf,     /* buffer to read into    */ 
                128,         /* size of read buffer    */ 
                &cNumRead) ) /* number of records read */ 
         sysError("ReadConsoleInput"); 
    }
 
    keyEvent = NULL;

    for (; nextEvent < cNumRead; nextEvent++)
    {
      if (irInBuf[nextEvent].EventType == KEY_EVENT)
      {
        keyEvent = &(irInBuf[nextEvent].Event.KeyEvent);
        if (keyEvent -> wRepeatCount-- <= 0) continue;
        if (!(keyEvent -> bKeyDown) && !listenRelease) continue;

        rk->map  = keyEvent->AsciiChar;
        rk->chr  = keyEvent->wVirtualScanCode;
        rk->fkey = keyEvent->wVirtualKeyCode;
        rk->c0   = keyEvent->dwControlKeyState & 255;
        rk->c1   = (keyEvent->dwControlKeyState >> 8) & 255;
        rk->c2   = !(keyEvent -> bKeyDown);

        if (rk->map > 0 && rk->map != 9) return;        // any key

        if (rk->map == 9) {                             // tab || back-tab
          if (rk->c0 & 16) rk->map = 29;  
          return;
        }

        // map == 0  -> special key
        // printf ("having ftion key\n");

        for (i = 0; moveKey [i] > 0; i++) { // do I recognize any spec. key?
          if (moveKey [i] == rk->fkey) {
            rk->map = mapMoveKey [i];           // recognized virtual key code
            return;
          }
        }
        rk -> map = 1;                          // other special keys

        for (i = 0; invisibleKey [i] >= 0; i++) {// test for invisible keys         
          if (rk->fkey == invisibleKey [i]) {   // like ctrl, shift, alt
            break;                              // and in that case
          }
        }
        if (invisibleKey [i] >= 0 && !listenCtrls) // process next key
          continue;
        return;                                 // else return this key
      }
      else // it was not key event (could be mouse event?)
      if (mouseEnabled && irInBuf[nextEvent].EventType == MOUSE_EVENT)
      {
        mouseEvent = &(irInBuf[nextEvent].Event.MouseEvent);
        rk->map  = 1;
        rk->chr  = 0;
        rk->fkey = (long)(mouseEvent->dwEventFlags) + 256;
        rk->c0   = mouseEvent->dwControlKeyState & 255;
        rk->c1   = (mouseEvent->dwControlKeyState >> 8) & 255;
        rk->c2   = mouseEvent -> dwButtonState & 255;
        rk->x    = (long)mouseEvent -> dwMousePosition.X + 1;
        rk->y    = (long)mouseEvent -> dwMousePosition.Y + 1;
        nextEvent++;
        return;
      }
    } // loop for all input records
  } // neverending loop
}

void EFlushInput ()
{
  flushed = TRUE;
  testOpened ("flushInput");
  FlushConsoleInputBuffer (hStdin);
}


void ElistenControllKeys (char enable)
{
  listenCtrls = enable;
}

void ElistenReleaseKeys (char enable)
{
  listenRelease = enable;
}

// fn looking for last nonspace and point behind
short nonSpaceLen (char const * t)
{
  int i = strlen (t); 
  while (--i >= 0 && t[i] == ' ');
  return ++i;
}

void insert (char what, char * where)
{
  short i = strlen (where) - 1;
  while (i > 0) where [i] = where [--i];
  where [0] = what;
}

// inline void EOUTTXT2(char * t, char a) { EOUTTXT ((t), strlen (t), (a)); }
#define EOUTTXT2(t, a)      EOUTTXT ((t), strlen (t), (a))

void EEDTTXT(char *fld, long w, long edtatr, long oldatr, void * pkt)
{
  keypkt * rk    = pkt;
  long  startLin = ECURLIN ();
  long  startPos = ECURPOS ();
  char  replaced;

  short aktPos;
  BOOL  curWasVisible = curinfo.bVisible;

  if (w > strlen (fld)) termError ("small buffer for editText");
  testOpened ("editText");

  aktPos = (insertMode ? nonSpaceLen (fld) : 0);   // cursor position

  replaced = fld [w] ;                // remember char under \000
  fld [w] = 0;                        // and set the end

  EOUTTXT2 (fld, edtatr);
  SetCursorPos (startPos + aktPos, startLin);
  rk -> map = 32;
  EshowCursor ();

  while (aktPos >= 0)
  {
    KEYIN (TRUE, rk);
    if (rk -> c2) continue;                           // continue for release
    if (rk -> map >= 32 && rk -> map != 127)
    {
      if (!insertMode) {                              // overwrite char
        fld [aktPos] = rk -> map;
        EOUTTXT2 (fld + aktPos, edtatr);
        if (aktPos < (w - 1)) aktPos++;
        SetCursorPos (startPos + aktPos, startLin);
        continue;
      } 
      else if (fld [w - 1] == ' ' && aktPos < w) {    // insert char
        insert (rk -> map, fld + aktPos);
        EOUTTXT2 (fld + aktPos, edtatr);
        aktPos++;
        SetCursorPos (startPos + aktPos, startLin);
        continue;
      }
    }
    else switch (rk -> map)
    {
      case 5: // left key
        if (aktPos > 0) {
          aktPos--;
          SetCursorPos (startPos + aktPos, startLin);
        }
        break;
      case 4: // right key
        if (aktPos < (w - !insertMode)) {
          aktPos++;
          SetCursorPos (startPos + aktPos, startLin);
        }
        break;
      case 127: // delete
        if (aktPos < w) {
          strcpy (fld + aktPos, fld + aktPos + 1);
          fld [w - 1] = ' ';
          EOUTTXT2 (fld + aktPos, edtatr);
          SetCursorPos (startPos + aktPos, startLin);
        }
        break;
      case 8: // backspace
        if (aktPos > 0) {
          strcpy (fld + aktPos - 1, fld + aktPos);
          fld [w - 1] = ' ';
          aktPos--;
          SetCursorPos (startPos + aktPos, startLin);
          EOUTTXT2 (fld + aktPos, edtatr);
          SetCursorPos (startPos + aktPos, startLin);
        }
        break;
      case 15: // home
        aktPos = 0;
        SetCursorPos (startPos + aktPos, startLin);
        break;
      case 26: // end
        aktPos = nonSpaceLen (fld);
        if (aktPos == w && !insertMode) aktPos--;
        SetCursorPos (startPos + aktPos, startLin);
        break;
      case 3:  // insert key
        setInsert (!insertMode);
        if (aktPos == w) SetCursorPos (startPos + (--aktPos), startLin);
        break;
      default:
        aktPos = -1;
    }
  }
  SetCursorPos (startPos, startLin);
  EOUTTXT2 (fld, oldatr);
  fld [w] = replaced;                // remembered char under \000

  if (!curWasVisible) EhideCursor ();
}

char EOPEN(long h, long w)
{
    SMALL_RECT sr = { 0, 0, 0, 0};
    COORD c;
    DWORD mode;
    long actH, actW;

    if (isOpened) {
      termError ("Trying to open opened console");
      return FALSE;
    }

    hStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
    if (hStdout == INVALID_HANDLE_VALUE) {
      sysError ("GetHandle - output"); 
      return FALSE;
    }

    GetConsoleScreenBufferInfo (hStdout, &screenInfo);

    actW = screenInfo.dwSize.X;
    actH = screenInfo.dwSize.Y;
    if (w != actW || h != actH) {
      if (w != actW)
        w = 80;
      if (h != actH && h != 50)
        h = 25;

      c.X = max (w, actW);
      c.Y = max (h, actH);
      if ((c.X != actW || c.Y != actH) 
          && !SetConsoleScreenBufferSize (hStdout, c))
        sysError ("cannot open term-buffer");
      sr.Right  = w - 1;
      sr.Bottom = h - 1;
      SetConsoleWindowInfo (hStdout, TRUE, &sr);
      if (c.X != w || c.Y != h)
      {
        c.X = w;
        c.Y = h;
        if (!SetConsoleScreenBufferSize (hStdout, c))
          sysError ("cannot open term-buffer");
      }
    }
    GetConsoleScreenBufferInfo (hStdout, &screenInfo);
    currAttr = screenInfo.wAttributes;

    if (! GetConsoleCursorInfo(hStdout, &curinfo))
      sysError ("GetConsoleCursorInfo");
    curSize = curinfo.dwSize;

    hStdin = GetStdHandle(STD_INPUT_HANDLE); 
    if (hStdin == INVALID_HANDLE_VALUE)
      sysError("GetStdHandle"); 

    // setting outmode:
    GetConsoleMode (hStdout, &mode);
    mode != ENABLE_WRAP_AT_EOL_OUTPUT;
    modeProcessed = (mode |= ENABLE_PROCESSED_OUTPUT);
    modeAll       = mode ^ ENABLE_PROCESSED_OUTPUT;
    printAll      = FALSE;
    SetConsoleMode (hStdout, mode);

    // setting of inmode:
    /* Enable the window (and mouse) input events. */
    // ENABLE_WINDOW_INPUT // | ENABLE_MOUSE_INPUT
    GetConsoleMode (hStdin, &oldInputMode);
/*
    mode = ((oldInputMode
             | ENABLE_PROCESSED_INPUT)
            & ~ENABLE_WINDOW_INPUT
            & ~ENABLE_MOUSE_INPUT
            & ~ENABLE_ECHO_INPUT	
            & ~ENABLE_LINE_INPUT);
*/
    mode = ENABLE_PROCESSED_INPUT | ENABLE_MOUSE_INPUT;
    if (! SetConsoleMode(hStdin, mode) )
        sysError("SetConsoleMode"); 

    mouseEnabled = TRUE;
    listenCtrls = listenRelease = FALSE;
 
    isOpened = TRUE;
    setInsert (FALSE);
    EFlushInput ();
    return TRUE;
}

void ECLOSE() 
{
    testOpened ("close");
//    if (! SetConsoleActiveScreenBuffer(hOldStdout) ||
//       ! CloseHandle (hStdout)) {
//      sysError("Closing console"); 
//    }
    if (! SetConsoleMode(hStdin, oldInputMode) )
    {
        sysError("Close");
    }
    setAttr(7);
    curinfo.bVisible = TRUE;
    curinfo.dwSize = curSize;
    if (! SetConsoleCursorInfo(hStdout, &curinfo))
      sysError ("Close, curInfo.");
    EUPDPOS (EHEIGHT (), 0);
    isOpened = FALSE;
    hStdout = INVALID_HANDLE_VALUE;
}

